"""add Series.tvdb_id

Revision ID: 572da51b5880
Revises: 016150597609
Create Date: 2019-02-16 13:28:02.033632

"""

# revision identifiers, used by Alembic.
revision = '572da51b5880'
down_revision = '016150597609'

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('series', sa.Column('tvdb_id', sa.Integer(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    pass



from golgi.config import configurable
from series.logging import Logging

from amino import IO

from series.api_client import ClientBase, ApiClientMeta, command, ApiClient


@configurable(library_client=['rest_api_port', 'rest_api_url'])
class LibClient(ClientBase, metaclass=ApiClientMeta):

    @property
    def client(self):
        return ApiClient(self._rest_api_url, self._rest_api_port)

    @command('series season episode', 'Create an episode with the supplied metadata')
    def create_episode(self, cmd):
        series, season, episode = cmd.args
        data = dict(episode=episode)
        path = 'series/{}/seasons/{}/episodes'.format(series, season)
        return IO.delay(self.client.post, path, body=data)

    @command('series season episode subfps', 'Set the episode\'s subfps')
    def subfps(self, cmd):
        series, season, episode, subfps = cmd.args
        data = dict(subfps=subfps)
        path = 'series/{}/seasons/{}/episodes/{}'.format(series, season,
                                                         episode)
        return IO.delay(self.client.put, path, body=data)

__all__ = ('LibClient',)



# -*- coding: utf-8 -*-

import os
import itertools
import re
import shutil
import operator
from pathlib import Path

import requests

from golgi import Config, cli, configurable
from tek.user_input import CheckboxList
from tek.tools import unicode_filename, sizeof_fmt, free_space_in_dir
from tek.errors import NotEnoughDiskSpace

from tek_utils import extract

from series import store_episode, subsync
from series.logging import Logging


@configurable(handle_episode=['archive_ext', 'episode_ext'])
class PathHandler(object):
    ext_re = '.*\.{}$'

    def __init__(self):
        self._archive_re = None
        self._episode_re = None

    def is_archive(self, fname):
        if not self._archive_re:
            self._archive_re = re.compile(
                self.ext_re.format(self._archive_ext)
            )
        return self._archive_re.match(fname) is not None

    def is_episode(self, fname):
        if not self._episode_re:
            self._episode_re = re.compile(
                self.ext_re.format(self._episode_ext)
            )
        return self._episode_re.match(str(fname)) is not None

    def find_episodes(self, location):
        location = Path(location)
        if location.is_file():
            candidates = [location]
        elif location.is_dir():
            candidates = self.find_files_recursive(location)
        else:
            return []
        return list(filter(self.is_episode, candidates))

    def find_files_recursive(self, basedir):
        for dir, dirs, files in os.walk(str(basedir)):
            for f in files:
                yield os.path.join(dir, f)


class File(Logging):
    def __init__(self, path):
        self.path = unicode_filename(str(path))

    @property
    def size(self):
        return os.path.getsize(self.path)

    @property
    def filename(self):
        return os.path.basename(self.path)

    def __repr__(self):
        return 'File({!r})'.format(self.path)

    def delete(self):
        os.remove(self.path)


@configurable(series=['temp_dir'])
class Episode(File):

    def __init__(self, source):
        self._source = source
        self._is_archive = isinstance(source, extract.ExtractJob)
        path = source.archive.path if self._is_archive else source
        self.missing_parts = (source.archive.missing_parts if self._is_archive
                              else None)
        File.__init__(self, path)
        self.files = []
        self.dest = None
        self._error = False
        self._source_file_in_temp = False
        self._path = PathHandler()

    def __repr__(self):
        return 'Episode({!r})'.format(self._source)

    def send_to_temp(self):
        if self._is_archive:
            self._extract_to_temp()
        else:
            self._file_to_temp()
        self.files = list(map(File, self._path.find_episodes(self.dest)))

    def _extract_to_temp(self):
        self.log.info('Extracting "{}" to temp dir…'.format(self.filename))
        self.dest = self._source.extract()
        self._error = bool(self._source.exitval)

    def _file_to_temp(self):
        dest = self._temp_dir / os.path.basename(self._source)
        if os.path.abspath(self._source) == Path(dest).absolute():
            self.log.info('File "{}" already in temp dir.'.format(self.filename))
            self._source_file_in_temp = True
        else:
            self._copy_to_temp()
        self.dest = dest.absolute()

    def _copy_to_temp(self):
        self.log.info('Copying "{}" to temp dir…'.format(self.filename))
        try:
            shutil.copy(str(self._source), str(self._temp_dir))
        except IOError as e:
            self._error = True
            self.log.error(e)

    @property
    def valid(self):
        return (not self._error and 0.95 * self.size <= self.target_size and
                not self._source_file_in_temp)

    @property
    def size(self):
        return (self._source.archive.size if self._is_archive else
                File.size.fget(self))

    @property
    def target_size(self):
        return sum([f.size for f in self.files])

    @property
    def size_string(self):
        return '{}: {} => {}'.format(self.filename, sizeof_fmt(self.size),
                                    sizeof_fmt(self.target_size))

    def remove_source(self):
        try:
            if self._is_archive:
                self._source.archive.delete()
            else:
                File.delete(self)
        except PermissionError as e:
            self.log.warning(e)

    @property
    def complete(self):
        return not self._is_archive or not self.missing_parts


def listdir_fullpath(dir):
    return list(map(os.path.join, itertools.repeat(dir),
                    os.listdir(dir)))


@configurable(series=['temp_dir', 'library_url'], handle_episode=['subtitles'])
class EpisodeHandler(Logging):

    def __init__(self, pretend_delete=False):
        self._pretend_delete = pretend_delete
        self._episodes = []
        self._path = PathHandler()

    def setup_episodes(self, sources):
        if not sources:
            sources = [os.getcwd()]
        dirs = filter(os.path.isdir, sources)
        files = list(itertools.filterfalse(os.path.isdir, sources))
        for d in dirs:
            files.extend(listdir_fullpath(d))
        files = [f for f in files if os.path.getsize(f) > 100]
        archives = filter(self._path.is_archive, files)
        archives = (extract.ExtractJob(a, dest_dir=str(self._temp_dir))
                    for a in archives)
        files = filter(self._path.is_episode, files)
        epi_files = set(itertools.chain(archives, files))
        self._episodes[:] = list(map(Episode, epi_files))
        total_size = sum([epi.size for epi in self._episodes], 0)
        available = free_space_in_dir(str(self._temp_dir))
        if total_size > available:
            raise NotEnoughDiskSpace(self._temp_dir, total_size, available)

    def handle_episodes(self, sources):
        try:
            self.setup_episodes(sources)
            for epi in self._episodes:
                if epi.complete:
                    epi.send_to_temp()
                else:
                    text = 'Episode "{}" is missing archive parts: {}'
                    parts = ', '.join(map(str, epi.missing_parts))
                    self.log.error(text.format(epi.path, parts))
            ag = operator.attrgetter
            input = CheckboxList(list(map(ag('size_string'), self._episodes)),
                                 list(map(ag('valid'), self._episodes)),
                                 text_post=['Remove source files?'])
            delete = input.read()
            if not self._pretend_delete:
                for epi, doit in zip(self._episodes, delete):
                    if doit:
                        epi.remove_source()
            if self._episode_files:
                storer = store_episode.store(self._episode_files)
                for epi in storer.stored:
                    url = '{}/series/{}/season/{}/episode/{}'
                    requests.post(url.format(self._library_url, epi.series,
                                             epi.season, epi.episode))

                def sync(epi):
                    if not epi.is_subtitle:
                        text = 'Downloading subs for episode {}…'
                        self.log.info(text.format(epi.filename))
                        return subsync.subsync_episode(epi, write=False)
                if self._subtitles:
                    subs = [_f for _f in map(sync, storer.stored) if _f]
                    if subs:
                        subsync.write_subs(subs)
                    else:
                        self.log.info('No subtitles found.')
        except NotEnoughDiskSpace as e:
            self.log.error(e)

    @property
    def _episode_files(self):
        attrs = lambda l, a: list(map(operator.attrgetter(a), l))
        return attrs(itertools.chain(*attrs(self._episodes, 'files')), 'path')


@cli(positional=('archive', '*'))
def handle_episode_cli():
    args = Config['extract'].archive
    EpisodeHandler().handle_episodes(args)



import operator
from typing import Tuple, Callable, TypeVar

from difflib import SequenceMatcher

from amino import do, Do, IO, Either, List, Try, Left, Maybe, Just, Nothing
from amino.json.data import JsonError
from amino.util.numeric import parse_int

from series.tvdb.client import query_list, query, query_one
from series.tvdb.data import (TvdbShow, TvdbSeason, TvdbSeasonSummary, TvdbEpisode, ShowCodec, SummaryCodec,
                              EpisodeCodec, TvdbEpisodeMetadata, EpisodeMetadataCodec)
from series.util import now_unix

A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')


def map_et(e: Either[A, B], f: Callable[[B], IO[Either[A, C]]]) -> IO[Either[A, C]]:
    return e.cata(lambda a: IO.pure(Left(a)), f)


@do(IO[Either[str, List[TvdbShow]]])
def search_show(name: str) -> Do:
    response = yield query_list(ShowCodec, 'search/series', name=name)
    return response.map(lambda data: data.map(TvdbShow.from_codec))


@do(IO[Either[JsonError, TvdbShow]])
def show_by_id(id: int) -> Do:
    response = yield query(ShowCodec, f'series/{id}')
    return response.map(TvdbShow.from_codec)


@do(Either[JsonError, TvdbSeasonSummary])
def analyse_summary(data: SummaryCodec) -> Do:
    num_epi = yield parse_int(data.airedEpisodes)
    seasons = yield data.airedSeasons.traverse(parse_int, Either)
    yield (
        Try(lambda: max(seasons))
        .map(lambda sc: TvdbSeasonSummary(num_epi, sc))
        .lmap(lambda a: JsonError(data, 'no seasons'))
    )


@do(IO[Either[JsonError, TvdbSeasonSummary]])
def season_summary(id: int) -> Do:
    response = yield query(SummaryCodec, f'series/{id}/episodes/summary')
    return response.flat_map(analyse_summary)


def season_result(number: int, response: List[EpisodeCodec]) -> TvdbSeason:
    return TvdbSeason(number, response.map(TvdbEpisode.from_codec))


@do(IO[Either[JsonError, TvdbSeason]])
def season(id: int, number: int) -> Do:
    response = yield query_list(EpisodeCodec, f'series/{id}/episodes/query', airedSeason=number)
    return response.map(lambda a: season_result(number, a))


@do(IO[Either[JsonError, TvdbSeason]])
def latest_season(id: int) -> Do:
    summary = yield season_summary(id)
    yield map_et(summary, lambda s: season(id, s.seasons))


CurrentEpisodes = Tuple[Maybe[TvdbEpisode], Maybe[TvdbEpisode]]


def closer(
        to_now: Callable[[int, int], bool],
        to_prev: Callable[[int, int], bool],
        now: int,
        prev: Maybe[TvdbEpisode],
) -> bool:
    def later(candidate: int) -> bool:
        prev_date = prev.flat_map(lambda a: a.airdate)
        return to_now(candidate, now) and not prev_date.exists(lambda a: to_prev(a, candidate))
    return later


def later(now: int, prev: Maybe[TvdbEpisode]) -> bool:
    return closer(operator.lt, operator.gt, now, prev)


def earlier(now: int, prev: Maybe[TvdbEpisode]) -> bool:
    return closer(operator.gt, operator.lt, now, prev)


def find_current_in_season(now: int, season: TvdbSeason) -> CurrentEpisodes:
    def folder(z: CurrentEpisodes, a: TvdbEpisode) -> CurrentEpisodes:
        prev_before, prev_after = z
        prev = Just(a) if a.airdate.exists(later(now, prev_before)) else prev_before
        after = Just(a) if a.airdate.exists(earlier(now, prev_after)) else prev_after
        return prev, after
    return season.episodes.fold_left((Nothing, Nothing))(folder)


@do(IO[Either[JsonError, CurrentEpisodes]])
def current_episodes(id: int) -> Do:
    season = yield latest_season(id)
    now = yield IO.delay(lambda: now_unix())
    return season.map(lambda a: find_current_in_season(now, a))


@do(IO[Either[JsonError, TvdbEpisode]])
def episode(id: int, sn: int, en: int) -> Do:
    data = yield query_one(EpisodeCodec, f'series/{id}/episodes/query', id=id, airedSeason=sn, airedEpisode=en)
    return data.map(TvdbEpisode.from_codec)


@do(IO[Either[JsonError, Maybe[int]]])
def airdate(id: int, sn: int, en: int) -> Do:
    response = yield episode(id, sn, en)
    return response.map(lambda a: a.airdate)


@do(IO[Either[JsonError, TvdbEpisodeMetadata]])
def episode_metadata(id: int, sn: int, en: int) -> Do:
    data = yield query_one(EpisodeMetadataCodec, f'series/{id}/episodes/query', id=id, airedSeason=sn, airedEpisode=en)
    return data.map(TvdbEpisodeMetadata.from_codec)


@do(IO[Either[JsonError, List[TvdbEpisodeMetadata]]])
def season_metadata(id: int, sn: int) -> Do:
    data = yield query_list(EpisodeMetadataCodec, f'series/{id}/episodes/query', airedSeason=sn)
    return data.map(lambda a: a.map(TvdbEpisodeMetadata.from_codec))


def name_match(target: str, candidate: TvdbShow) -> float:
    matcher = SequenceMatcher()
    matcher.set_seq2(target.lower())
    matcher.set_seq1(candidate.name.lower())
    return matcher.ratio()


@do(IO[Either[JsonError, Maybe[TvdbShow]]])
def guess_show(name: str) -> Do:
    candidates = yield search_show(name)
    return candidates.map(lambda a: a.sort_by(lambda b: name_match(name, b)).last)


__all__ = ('search_show', 'show_by_id', 'season', 'latest_season', 'current_episodes', 'airdate', 'episode_metadata',
           'season_metadata',)



from datetime import datetime

from series.get.handler import R, ProcessHandler
from series.condition import LambdaCondition
from series.get.search import TorrentSearch, SearchResults, SearchQuery
from series.util import now_unix
from series.handler import Job

from golgi.config import configurable

from tek_utils.sharehoster.torrent import SearchResult

from amino import __, Just, _, L, List, Maybe


@configurable(torrent=['search_engine'], get=['torrent_recheck_interval',
                                              'min_seeders', 'max_size'])
class TorrentFinder(ProcessHandler):

    def __init__(self, releases, *a, **kw):
        super().__init__(2, releases, 5, 'torrent finder',
                         cooldown=self._torrent_recheck_interval, **kw)

    def _create_async(self, monitor):
        self.log.info('Torrent search: {}'.format(monitor.release))
        self._update(monitor, last_torrent_search=datetime.now(), activated=False)
        return Just(TorrentSearch(monitor, self._description,
                                  self._search_engine, self._min_seeders,
                                  self._max_size))

    def _clean_done(self, a):
        p = a.proc
        if p.done.value:
            monitor = self._releases.find_by_id(p.monitor.id)
            a.proc.result % __.find(L(self._process_results)(monitor, _))
            return True

    def _clean_timeout(self, proc):
        pass

    def _process_results(self, monitor, results: SearchResults):
        ''' returns True if a result was found to stop the find() in
        *_clean_done*
        '''
        return (
            results.choose(monitor)
            .map(L(self._add_link)(results.query, _))
            .replace(True)
            .get_or_else(L(self._no_result)(results))
        )

    def _add_link(self, query: SearchQuery, result: SearchResult):
        self.log.info('Added torrent to release {}: {} ({} seeders)'
                      .format(query.release, result.title, result.seeders))
        self._releases.add_torrent_by_id(query.monitor.id, result.magnet_link)
        self._update(query.monitor, last_torrent_update_stamp=now_unix())

    def _no_result(self, results):
        self.log.debug('None of the results match the release.')
        self.log.debug('min_seeders: {}'.format(self._min_seeders))
        self.log.debug('\n'.join([r.title for r in results.results]))

    @property
    def _conditions(self):
        return (
            ~R('downloaded') & ~R('has_cachable_torrents') &
            LambdaCondition('recheck interval',
                            __.can_recheck(self._torrent_recheck_interval))
        )

    def _choose(self, qualified: List[Job]) -> Maybe[Job]:
        return qualified.find(_.item.activated).o(qualified.head)

    @property
    def _candidates(self):
        return self._no_cached_torrents

    def activate_id(self, id):
        super().activate_id(id)
        self._releases.update_by_id(id, last_torrent_search_stamp=0, activated=True)

    @property
    def status(self):
        return self._async / _.monitor.release.info / List | List()

__all__ = ['TorrentFinder']



import threading
import time
import importlib

from golgi.run import SignalManager

from amino import List, Map, Maybe
from amino.util.string import camelcaseify

from series.logging import Logging
from series.handler import Handler


class App(threading.Thread, Logging):
    _components = []  # type: List[Handler]

    def __init__(self, c_module, run, omit, c_args=(), interval=0.2,
                 name='series'):
        super().__init__(name=name)
        self._c_module = c_module
        self._c_args = c_args
        self._interval = interval
        self.name = name
        self._running = False
        self.components = List()
        self.component_map = Map()
        self._init_components(run, omit)

    def run(self):
        self.register_sigint_handler()
        self.pre()
        self.main_loop()

    def pre(self):
        self.start_components()
        self.prepare()

    def prepare(self):
        pass

    def start_components(self):
        for component in self.components:
            component.start()

    def main_loop(self):
        self._running = True
        self.log.info('{} running.'.format(self.name))
        while self._running:
            time.sleep(self._interval)

    def _init_components(self, run, omit):
        components = self._components
        if run:
            components = [c for c in components if c in run]
        if omit:
            components = [c for c in components if c not in omit]
        for component in components:
            self._setup_component(component)

    def _setup_component(self, component):
        def error(exc):
            msg = 'Invalid component name: {} ({})'.format(component, exc)
            self.log.error(msg)
        mod_name = 'series.{}.{}'.format(self._c_module, component)
        try:
            module = importlib.import_module(mod_name)
        except ImportError as e:
            error(e)
        else:
            try:
                Component = getattr(module, camelcaseify(component))
            except AttributeError as e:
                error(e)
            else:
                instance = Component(*self._c_args)
                setattr(self, component, instance)
                self.components.append(instance)
                self.component_map[component] = instance

    def register_sigint_handler(self):
        SignalManager.instance.sigint(self.interrupt)

    def interrupt(self, signum=10, frame=None):
        self._stop_components()
        self._join_threads()
        self._cleanup()
        self._running = False
        self.log.info('{} shut down due to SIGINT.'.format(self.name))

    def _stop_components(self):
        for component in self.components:
            component.stop()

    def _join_threads(self):
        for component in self.components:
            if component.is_alive():
                component.join()

    def _cleanup(self):
        pass

    def component(self, name: str) -> Maybe[Handler]:
        return self.component_map.lift(name)

__all__ = ['App']



import time
import threading
from datetime import datetime

from tek.tools import wait_for

from series.library.model.episode import Episode


class PlaybackObserver(threading.Thread):

    def __init__(self, mplayer, library, video, player):
        super().__init__()
        self._mplayer = mplayer
        self._library = library
        self._video = video
        self._player = player
        self._start_time = None
        self._finish_time = None
        self._stopped = False
        self._last_pos = None

    def run(self):
        wait_for(lambda: self._player.running)
        self._start_time = datetime.now()
        while self._running:
            self._set_position()
            time.sleep(1)
        self._finish()

    def stop(self):
        self._stopped = True

    def _set_position(self):
        position = self._player.elapsed
        if position:
            self._last_pos = position

    def _finish(self):
        self._finish_time = datetime.now()
        if isinstance(self._video, Episode):
            self._library.add_watch_event(
                self._video.series,
                self._video.season,
                self._video.number,
                self._start_time,
                self._finish_time,
                self._last_pos
            )

    @property
    def _running(self):
        ''' Test whether the same video is still loaded in mplayer.
        Either it has been terminated manually, which means that stop()
        has been called, or the video is at its end, in which case the
        'running' attribute of the player and the current observer is
        used.
        '''
        return (not self._stopped and self._player.running and
                self._player.same_episode(self))

__all__ = ['PlaybackObserver']



"""add ReleaseMonitor.airdate_stamp

Revision ID: 1d11afd9a90
Revises: 2ee6e71eb2f
Create Date: 2014-12-18 11:43:35.199129

"""

# revision identifiers, used by Alembic.
revision = '1d11afd9a90'
down_revision = '2ee6e71eb2f'

from alembic import op
import sqlalchemy as sa


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.add_column('release', sa.Column('airdate_stamp', sa.Integer(), nullable=True))
    ### end Alembic commands ###


def downgrade():
    pass



import requests

from amino import IO, Map, Either, List, Right, Left, do, Do, Dat, Maybe, Nil
from amino.json.decoder import decode_json_type, decode_json_type_json
from amino.json.data import Json, JsonError

from typing import TypeVar, Type, Callable, Any, Dict, Generic

A = TypeVar('A')
P = TypeVar('P')
api_key = 'Y08E38CQDT92BQ5T'
auth_token = None
base_url = 'https://api.thetvdb.com'


def url(path: str) -> str:
    return f'{base_url}/{path}'


@do(IO[None])
def auth() -> Do:
    response = yield IO.delay(requests.post, url('login'), json=Map(apikey=api_key))
    body = yield IO.delay(lambda: response.json())
    token_field = yield (
        IO.failed(f'invalid response payload structure from the tvdb api auth endpoint: {body}')
        if not isinstance(body, dict) else
        IO.pure(Map(body).lift('token'))
    )
    token = yield IO.from_maybe(token_field, f'no token in the tvdb api auth response: {body}')
    global auth_token
    auth_token = token


def headers() -> dict:
    return {
        'accept-language': 'en',
        'authorization': f'Bearer {auth_token}',
    }


@do(IO[str])
def request(path: str, params: Map[str, str]) -> Do:
    response = yield IO.delay(requests.get, url(path), params=params, headers=headers())
    status = response.status_code
    if status == 401:
        yield auth()
        yield request(path, params)
    elif status == 200:
        yield IO.delay(lambda: response.text)
    else:
        yield IO.failed(f'unexpected response code {status} from the tvdb api for {path}')


@do(IO[Either[JsonError, A]])
def data_query(
        path: str,
        params: Map[str, str],
        payload_tpe: type,
        decode_data: Callable[[Any], Either[JsonError, A]],
) -> Do:
    response = yield request(path, params)
    data = decode_json_type(response, payload_tpe)
    return data.flat_map(decode_data)


class Links(Dat['Links']):

    def __init__(self, next: Maybe[int]) -> None:
        self.next = next


class JsonPayload(Dat['JsonPayload']):

    def __init__(self, data: Json) -> None:
        self.data = data


def query(tpe: Type[A], path: str, **params: Dict[str, str]) -> IO[Either[JsonError, A]]:
    return data_query(path, Map(params), JsonPayload, lambda a: decode_json_type_json(a.data, tpe))


class JsonListPayload(Dat['JsonListPayload']):

    def __init__(self, data: List[Json], links: Maybe[Links]) -> None:
        self.data = data
        self.links = links


class ListResult(Generic[A], Dat['ListResult[A]']):

    def __init__(self, data: List[A], next: Maybe[int]) -> None:
        self.data = data
        self.next = next


def decode_list(tpe: Type[A]) -> Callable[[JsonListPayload], Either[JsonError, ListResult[A]]]:
    @do(Either[JsonError, ListResult[A]])
    def decode_list(payload: JsonListPayload) -> Do:
        data = yield payload.data.traverse(lambda a: decode_json_type_json(a, tpe), Either)
        return ListResult(data, payload.links.flat_map(lambda a: a.next))
    return decode_list


@do(IO[Either[JsonError, List[A]]])
def query_list(tpe: Type[A], path: str, **params: Dict[str, str]) -> Do:
    response = yield data_query(
        path,
        Map(params),
        JsonListPayload,
        decode_list(tpe),
    )
    @do(IO[Either[JsonError, List[A]]])
    def unpack_and_recurse(result: ListResult[A]) -> Do:
        next_data = yield (
            result.next
            .map(lambda a: query_list(tpe, path, page=a, **params))
            .get_or_strict(IO.pure(Right(Nil)))
        )
        return next_data.map(lambda a: result.data + a)
    yield response.cata(lambda a: IO.pure(Left(a)), unpack_and_recurse)


@do(Either[JsonError, A])
def validate_query_one(result: Either[JsonError, List[A]], params: Dict[str, str]) -> Do:
    items = yield result
    err = lambda what: JsonError(items, f'{what} matched for {params}')
    def multi() -> Either[JsonError, A]:
        return Left(err('multiple records'))
    yield (
        items
        .detach_head
        .to_either(lambda: err(items, 'no record'))
        .flat_map2(lambda h, t: Right(h) if t.empty else multi())
    )


@do(IO[Either[JsonError, A]])
def query_one(tpe: Type[A], path: str, **params: Dict[str, str]) -> Do:
    items = yield query_list(tpe, path, **params)
    return validate_query_one(items, params)


__all__ = ('api_key', 'auth', 'query', 'query_list', 'query_one',)



"""add Episode.subdelay

Revision ID: 016150597609
Revises: 1f69c57af9a
Create Date: 2017-06-24 20:26:51.272430

"""

# revision identifiers, used by Alembic.
revision = '016150597609'
down_revision = '1f69c57af9a'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('episode', sa.Column('subdelay', sa.Float(), nullable=True))


def downgrade():
    pass
